import * as path from 'path';
import * as vscode from 'vscode';
import { AvatarManager } from './avatarManager';
import { getConfig } from './config';
import { DataSource, GIT_CONFIG_USER_EMAIL, GIT_CONFIG_USER_NAME, GitCommitDetailsData } from './dataSource';
import { ExtensionState } from './extensionState';
import { Logger } from './logger';
import { RepoFileWatcher } from './repoFileWatcher';
import { RepoManager } from './repoManager';
import { ErrorInfo, GitConfigLocation, GitGraphViewInitialState, GitRepoSet, LoadGitGraphViewTo, RefLabelAlignment, RequestMessage, ResponseMessage, TabIconColourTheme } from './types';
import { copyFilePathToClipboard, copyToClipboard, getNonce, openExtensionSettings, openFile, showErrorMessage, UNABLE_TO_FIND_GIT_MSG, UNCOMMITTED, viewDiff, viewScm } from './utils';

export class GitGraphView {
	public static currentPanel: GitGraphView | undefined;

	private readonly panel: vscode.WebviewPanel;
	private readonly extensionPath: string;
	private readonly avatarManager: AvatarManager;
	private readonly dataSource: DataSource;
	private readonly extensionState: ExtensionState;
	private readonly repoFileWatcher: RepoFileWatcher;
	private readonly repoManager: RepoManager;
	private readonly logger: Logger;
	private disposables: vscode.Disposable[] = [];
	private isGraphViewLoaded: boolean = false;
	private isPanelVisible: boolean = true;
	private currentRepo: string | null = null;
	private loadViewTo: LoadGitGraphViewTo = null; // Is used by the next call to getHtmlForWebview, and is then reset to null

	private loadRepoInfoRefreshId: number = 0;
	private loadCommitsRefreshId: number = 0;

	public static createOrShow(extensionPath: string, dataSource: DataSource, extensionState: ExtensionState, avatarManager: AvatarManager, repoManager: RepoManager, logger: Logger, loadViewTo: LoadGitGraphViewTo) {
		const column = vscode.window.activeTextEditor ? vscode.window.activeTextEditor.viewColumn : undefined;

		if (GitGraphView.currentPanel) {
			// If Git Graph panel already exists
			if (GitGraphView.currentPanel.isPanelVisible) {
				// If the Git Graph panel is visible
				if (loadViewTo !== null) {
					GitGraphView.currentPanel.respondLoadRepos(repoManager.getRepos(), loadViewTo);
				}
			} else {
				// If the Git Graph panel is not visible 
				GitGraphView.currentPanel.loadViewTo = loadViewTo;
			}
			GitGraphView.currentPanel.panel.reveal(column);
		} else {
			// If Git Graph panel doesn't already exist
			GitGraphView.currentPanel = new GitGraphView(extensionPath, dataSource, extensionState, avatarManager, repoManager, logger, loadViewTo, column);
		}
	}

	private constructor(extensionPath: string, dataSource: DataSource, extensionState: ExtensionState, avatarManager: AvatarManager, repoManager: RepoManager, logger: Logger, loadViewTo: LoadGitGraphViewTo, column: vscode.ViewColumn | undefined) {
		this.extensionPath = extensionPath;
		this.avatarManager = avatarManager;
		this.dataSource = dataSource;
		this.extensionState = extensionState;
		this.repoManager = repoManager;
		this.logger = logger;
		this.loadViewTo = loadViewTo;
		this.avatarManager.registerView(this);

		const config = getConfig();
		this.panel = vscode.window.createWebviewPanel('git-graph', 'Git Graph', column || vscode.ViewColumn.One, {
			enableScripts: true,
			localResourceRoots: [vscode.Uri.file(path.join(extensionPath, 'media'))],
			retainContextWhenHidden: config.retainContextWhenHidden
		});
		this.panel.iconPath = config.tabIconColourTheme === TabIconColourTheme.Colour
			? this.getUri('resources', 'webview-icon.svg')
			: { light: this.getUri('resources', 'webview-icon-light.svg'), dark: this.getUri('resources', 'webview-icon-dark.svg') };

		this.update();
		this.panel.onDidDispose(() => this.dispose(), null, this.disposables);
		this.panel.onDidChangeViewState(() => {
			if (this.panel.visible !== this.isPanelVisible) {
				if (this.panel.visible) {
					this.update();
				} else {
					this.currentRepo = null;
					this.repoFileWatcher.stop();
				}
				this.isPanelVisible = this.panel.visible;
			}
		}, null, this.disposables);

		this.repoFileWatcher = new RepoFileWatcher(logger, () => {
			if (this.panel.visible) {
				this.sendMessage({ command: 'refresh' });
			}
		});
		this.repoManager.registerViewCallback((repos: GitRepoSet, numRepos: number, loadRepo: string | null) => {
			if (!this.panel.visible) return;
			const loadViewTo = loadRepo !== null ? { repo: loadRepo, commitDetails: null } : null;
			if ((numRepos === 0 && this.isGraphViewLoaded) || (numRepos > 0 && !this.isGraphViewLoaded)) {
				this.loadViewTo = loadViewTo;
				this.update();
			} else {
				this.respondLoadRepos(repos, loadViewTo);
			}
		});

		this.panel.webview.onDidReceiveMessage(async (msg: RequestMessage) => {
			if (this.dataSource === null) return;
			this.repoFileWatcher.mute();
			let errorInfos: ErrorInfo[];

			switch (msg.command) {
				case 'addRemote':
					this.sendMessage({
						command: 'addRemote',
						error: await this.dataSource.addRemote(msg.repo, msg.name, msg.url, msg.pushUrl, msg.fetch)
					});
					break;
				case 'addTag':
					errorInfos = [await this.dataSource.addTag(msg.repo, msg.tagName, msg.commitHash, msg.lightweight, msg.message)];
					if (errorInfos[0] === null && msg.pushToRemote !== null) {
						errorInfos.push(await this.dataSource.pushTag(msg.repo, msg.tagName, msg.pushToRemote));
					}
					this.sendMessage({ command: 'addTag', errors: errorInfos });
					break;
				case 'applyStash':
					this.sendMessage({
						command: 'applyStash',
						error: await this.dataSource.applyStash(msg.repo, msg.selector, msg.reinstateIndex)
					});
					break;
				case 'branchFromStash':
					this.sendMessage({
						command: 'branchFromStash',
						error: await this.dataSource.branchFromStash(msg.repo, msg.selector, msg.branchName)
					});
					break;
				case 'checkoutBranch':
					this.sendMessage({
						command: 'checkoutBranch',
						error: await this.dataSource.checkoutBranch(msg.repo, msg.branchName, msg.remoteBranch)
					});
					break;
				case 'checkoutCommit':
					this.sendMessage({
						command: 'checkoutCommit',
						error: await this.dataSource.checkoutCommit(msg.repo, msg.commitHash)
					});
					break;
				case 'cherrypickCommit':
					errorInfos = [await this.dataSource.cherrypickCommit(msg.repo, msg.commitHash, msg.parentIndex, msg.recordOrigin, msg.noCommit)];
					if (errorInfos[0] === null && msg.noCommit) {
						errorInfos.push(await viewScm());
					}
					this.sendMessage({ command: 'cherrypickCommit', errors: errorInfos });
					break;
				case 'cleanUntrackedFiles':
					this.sendMessage({
						command: 'cleanUntrackedFiles',
						error: await this.dataSource.cleanUntrackedFiles(msg.repo, msg.directories)
					});
					break;
				case 'codeReviewFileReviewed':
					this.extensionState.updateCodeReviewFileReviewed(msg.repo, msg.id, msg.filePath);
					break;
				case 'commitDetails':
					let data = await Promise.all<GitCommitDetailsData, string | null>([
						msg.commitHash === UNCOMMITTED
							? this.dataSource.getUncommittedDetails(msg.repo)
							: msg.stash === null
								? this.dataSource.getCommitDetails(msg.repo, msg.commitHash)
								: this.dataSource.getStashDetails(msg.repo, msg.commitHash, msg.stash),
						msg.avatarEmail !== null ? this.avatarManager.getAvatarImage(msg.avatarEmail) : Promise.resolve(null)
					]);
					this.sendMessage({
						command: 'commitDetails',
						...data[0],
						avatar: data[1],
						codeReview: msg.commitHash !== UNCOMMITTED ? this.extensionState.getCodeReview(msg.repo, msg.commitHash) : null,
						refresh: msg.refresh
					});
					break;
				case 'compareCommits':
					this.sendMessage({
						command: 'compareCommits',
						commitHash: msg.commitHash, compareWithHash: msg.compareWithHash,
						... await this.dataSource.getCommitComparison(msg.repo, msg.fromHash, msg.toHash),
						codeReview: msg.toHash !== UNCOMMITTED ? this.extensionState.getCodeReview(msg.repo, msg.fromHash + '-' + msg.toHash) : null,
						refresh: msg.refresh
					});
					break;
				case 'copyFilePath':
					this.sendMessage({
						command: 'copyFilePath',
						error: await copyFilePathToClipboard(msg.repo, msg.filePath)
					});
					break;
				case 'copyToClipboard':
					this.sendMessage({
						command: 'copyToClipboard',
						type: msg.type,
						error: await copyToClipboard(msg.data)
					});
					break;
				case 'createBranch':
					this.sendMessage({
						command: 'createBranch',
						error: await this.dataSource.createBranch(msg.repo, msg.branchName, msg.commitHash, msg.checkout)
					});
					break;
				case 'deleteBranch':
					errorInfos = [await this.dataSource.deleteBranch(msg.repo, msg.branchName, msg.forceDelete)];
					for (let i = 0; i < msg.deleteOnRemotes.length; i++) {
						errorInfos.push(await this.dataSource.deleteRemoteBranch(msg.repo, msg.branchName, msg.deleteOnRemotes[i]));
					}
					this.sendMessage({ command: 'deleteBranch', errors: errorInfos });
					break;
				case 'deleteRemote':
					this.sendMessage({
						command: 'deleteRemote',
						error: await this.dataSource.deleteRemote(msg.repo, msg.name)
					});
					break;
				case 'deleteRemoteBranch':
					this.sendMessage({
						command: 'deleteRemoteBranch',
						error: await this.dataSource.deleteRemoteBranch(msg.repo, msg.branchName, msg.remote)
					});
					break;
				case 'deleteTag':
					this.sendMessage({
						command: 'deleteTag',
						error: await this.dataSource.deleteTag(msg.repo, msg.tagName, msg.deleteOnRemote)
					});
					break;
				case 'deleteUserDetails':
					errorInfos = [];
					if (msg.name) {
						errorInfos.push(await this.dataSource.unsetConfigValue(msg.repo, GIT_CONFIG_USER_NAME, msg.location));
					}
					if (msg.email) {
						errorInfos.push(await this.dataSource.unsetConfigValue(msg.repo, GIT_CONFIG_USER_EMAIL, msg.location));
					}
					this.sendMessage({
						command: 'deleteUserDetails',
						errors: errorInfos
					});
					break;
				case 'dropCommit':
					this.sendMessage({
						command: 'dropCommit',
						error: await this.dataSource.dropCommit(msg.repo, msg.commitHash)
					});
					break;
				case 'dropStash':
					this.sendMessage({
						command: 'dropStash',
						error: await this.dataSource.dropStash(msg.repo, msg.selector)
					});
					break;
				case 'editRemote':
					this.sendMessage({
						command: 'editRemote',
						error: await this.dataSource.editRemote(msg.repo, msg.nameOld, msg.nameNew, msg.urlOld, msg.urlNew, msg.pushUrlOld, msg.pushUrlNew)
					});
					break;
				case 'editUserDetails':
					errorInfos = [
						await this.dataSource.setConfigValue(msg.repo, GIT_CONFIG_USER_NAME, msg.name, msg.location),
						await this.dataSource.setConfigValue(msg.repo, GIT_CONFIG_USER_EMAIL, msg.email, msg.location)
					];
					if (errorInfos[0] === null && errorInfos[1] === null) {
						if (msg.deleteLocalName) {
							errorInfos.push(await this.dataSource.unsetConfigValue(msg.repo, GIT_CONFIG_USER_NAME, GitConfigLocation.Local));
						}
						if (msg.deleteLocalEmail) {
							errorInfos.push(await this.dataSource.unsetConfigValue(msg.repo, GIT_CONFIG_USER_EMAIL, GitConfigLocation.Local));
						}
					}
					this.sendMessage({
						command: 'editUserDetails',
						errors: errorInfos
					});
					break;
				case 'fetch':
					this.sendMessage({
						command: 'fetch',
						error: await this.dataSource.fetch(msg.repo, msg.name, msg.prune)
					});
					break;
				case 'fetchAvatar':
					this.avatarManager.fetchAvatarImage(msg.email, msg.repo, msg.remote, msg.commits);
					break;
				case 'fetchIntoLocalBranch':
					this.sendMessage({
						command: 'fetchIntoLocalBranch',
						error: await this.dataSource.fetchIntoLocalBranch(msg.repo, msg.remote, msg.remoteBranch, msg.localBranch)
					});
					break;
				case 'endCodeReview':
					this.extensionState.endCodeReview(msg.repo, msg.id);
					break;
				case 'getSettings':
					this.sendMessage({
						command: 'getSettings',
						... await this.dataSource.getRepoSettings(msg.repo)
					});
					break;
				case 'loadCommits':
					this.loadCommitsRefreshId = msg.refreshId;
					this.sendMessage({
						command: 'loadCommits',
						refreshId: msg.refreshId,
						... await this.dataSource.getCommits(msg.repo, msg.branches, msg.maxCommits, msg.showRemoteBranches, msg.showTags, msg.remotes, msg.hideRemotes)
					});
					break;
				case 'loadRepoInfo':
					this.loadRepoInfoRefreshId = msg.refreshId;
					let repoInfo = await this.dataSource.getRepoInfo(msg.repo, msg.showRemoteBranches, msg.hideRemotes), isRepo = true;
					if (repoInfo.error) {
						// If an error occurred, check to make sure the repo still exists
						isRepo = (await this.dataSource.repoRoot(msg.repo)) !== null;
						if (!isRepo) repoInfo.error = null; // If the error is caused by the repo no longer existing, clear the error message
					}
					this.sendMessage({
						command: 'loadRepoInfo',
						refreshId: msg.refreshId,
						...repoInfo,
						isRepo: isRepo
					});
					if (msg.repo !== this.currentRepo) {
						this.currentRepo = msg.repo;
						this.extensionState.setLastActiveRepo(msg.repo);
						this.repoFileWatcher.start(msg.repo);
					}
					break;
				case 'loadRepos':
					if (!msg.check || !await this.repoManager.checkReposExist()) {
						// If not required to check repos, or no changes were found when checking, respond with repos
						this.respondLoadRepos(this.repoManager.getRepos(), null);
					}
					break;
				case 'merge':
					this.sendMessage({
						command: 'merge', actionOn: msg.actionOn,
						error: await this.dataSource.merge(msg.repo, msg.obj, msg.actionOn, msg.createNewCommit, msg.squash, msg.noCommit)
					});
					break;
				case 'openExtensionSettings':
					this.sendMessage({
						command: 'openExtensionSettings',
						error: await openExtensionSettings()
					});
					break;
				case 'openFile':
					this.sendMessage({
						command: 'openFile',
						error: await openFile(msg.repo, msg.filePath)
					});
					break;
				case 'popStash':
					this.sendMessage({
						command: 'popStash',
						error: await this.dataSource.popStash(msg.repo, msg.selector, msg.reinstateIndex)
					});
					break;
				case 'pruneRemote':
					this.sendMessage({
						command: 'pruneRemote',
						error: await this.dataSource.pruneRemote(msg.repo, msg.name)
					});
					break;
				case 'pullBranch':
					this.sendMessage({
						command: 'pullBranch',
						error: await this.dataSource.pullBranch(msg.repo, msg.branchName, msg.remote, msg.createNewCommit, msg.squash)
					});
					break;
				case 'pushBranch':
					this.sendMessage({
						command: 'pushBranch',
						error: await this.dataSource.pushBranch(msg.repo, msg.branchName, msg.remote, msg.setUpstream, msg.mode)
					});
					break;
				case 'pushStash':
					this.sendMessage({
						command: 'pushStash',
						error: await this.dataSource.pushStash(msg.repo, msg.message, msg.includeUntracked)
					});
					break;
				case 'pushTag':
					this.sendMessage({
						command: 'pushTag',
						error: await this.dataSource.pushTag(msg.repo, msg.tagName, msg.remote)
					});
					break;
				case 'rebase':
					this.sendMessage({
						command: 'rebase', actionOn: msg.actionOn, interactive: msg.interactive,
						error: await this.dataSource.rebase(msg.repo, msg.obj, msg.actionOn, msg.ignoreDate, msg.interactive)
					});
					break;
				case 'renameBranch':
					this.sendMessage({
						command: 'renameBranch',
						error: await this.dataSource.renameBranch(msg.repo, msg.oldName, msg.newName)
					});
					break;
				case 'rescanForRepos':
					if (!(await this.repoManager.searchWorkspaceForRepos())) {
						showErrorMessage('No Git repositories were found in the current workspace.');
					}
					break;
				case 'resetToCommit':
					this.sendMessage({
						command: 'resetToCommit',
						error: await this.dataSource.resetToCommit(msg.repo, msg.commit, msg.resetMode)
					});
					break;
				case 'revertCommit':
					this.sendMessage({
						command: 'revertCommit',
						error: await this.dataSource.revertCommit(msg.repo, msg.commitHash, msg.parentIndex)
					});
					break;
				case 'setGlobalViewState':
					this.sendMessage({
						command: 'setGlobalViewState',
						error: await this.extensionState.setGlobalViewState(msg.state)
					});
					break;
				case 'setRepoState':
					this.repoManager.setRepoState(msg.repo, msg.state);
					break;
				case 'showErrorMessage':
					showErrorMessage(msg.message);
					break;
				case 'startCodeReview':
					this.sendMessage({
						command: 'startCodeReview',
						commitHash: msg.commitHash,
						compareWithHash: msg.compareWithHash,
						... await this.extensionState.startCodeReview(msg.repo, msg.id, msg.files, msg.lastViewedFile)
					});
					break;
				case 'tagDetails':
					this.sendMessage({
						command: 'tagDetails',
						tagName: msg.tagName,
						commitHash: msg.commitHash,
						... await this.dataSource.getTagDetails(msg.repo, msg.tagName)
					});
					break;
				case 'viewDiff':
					this.sendMessage({
						command: 'viewDiff',
						error: await viewDiff(msg.repo, msg.fromHash, msg.toHash, msg.oldFilePath, msg.newFilePath, msg.type)
					});
					break;
				case 'viewScm':
					this.sendMessage({
						command: 'viewScm',
						error: await viewScm()
					});
					break;
			}

			this.repoFileWatcher.unmute();
		}, null, this.disposables);

		this.logger.log('Created Git Graph View' + (loadViewTo !== null ? ' (active repo: ' + loadViewTo.repo + ')' : ''));
	}

	public sendMessage(msg: ResponseMessage) {
		this.panel.webview.postMessage(msg);
	}

	public dispose() {
		GitGraphView.currentPanel = undefined;
		this.panel.dispose();
		this.avatarManager.deregisterView();
		this.repoFileWatcher.stop();
		this.repoManager.deregisterViewCallback();
		while (this.disposables.length) {
			const x = this.disposables.pop();
			if (x) x.dispose();
		}
		this.logger.log('Disposed Git Graph View');
	}

	private update() {
		this.panel.webview.html = this.getHtmlForWebview();
	}

	private getHtmlForWebview() {
		const config = getConfig(), nonce = getNonce();
		const refLabelAlignment = config.refLabelAlignment;
		const initialState: GitGraphViewInitialState = {
			config: {
				autoCenterCommitDetailsView: config.autoCenterCommitDetailsView,
				branchLabelsAlignedToGraph: refLabelAlignment === RefLabelAlignment.BranchesAlignedToGraphAndTagsOnRight,
				combineLocalAndRemoteBranchLabels: config.combineLocalAndRemoteBranchLabels,
				commitDetailsViewLocation: config.commitDetailsViewLocation,
				contextMenuActionsVisibility: config.contextMenuActionsVisibility,
				customBranchGlobPatterns: config.customBranchGlobPatterns,
				customEmojiShortcodeMappings: config.customEmojiShortcodeMappings,
				dateFormat: config.dateFormat,
				defaultColumnVisibility: config.defaultColumnVisibility,
				defaultFileViewType: config.defaultFileViewType,
				dialogDefaults: config.dialogDefaults,
				fetchAndPrune: config.fetchAndPrune,
				fetchAvatars: config.fetchAvatars && this.extensionState.isAvatarStorageAvailable(),
				graphColours: config.graphColours,
				graphStyle: config.graphStyle,
				grid: { x: 16, y: 24, offsetX: 16, offsetY: 12, expandY: 250 },
				initialLoadCommits: config.initialLoadCommits,
				loadMoreCommits: config.loadMoreCommits,
				loadMoreCommitsAutomatically: config.loadMoreCommitsAutomatically,
				muteCommitsNotAncestorsOfHead: config.muteCommitsThatAreNotAncestorsOfHead,
				muteMergeCommits: config.muteMergeCommits,
				openRepoToHead: config.openRepoToHead,
				showCurrentBranchByDefault: config.showCurrentBranchByDefault,
				showTags: config.showTags,
				tagLabelsOnRight: refLabelAlignment !== RefLabelAlignment.Normal
			},
			lastActiveRepo: this.extensionState.getLastActiveRepo(),
			loadViewTo: this.loadViewTo,
			repos: this.repoManager.getRepos(),
			loadRepoInfoRefreshId: this.loadRepoInfoRefreshId,
			loadCommitsRefreshId: this.loadCommitsRefreshId
		};
		const globalState = this.extensionState.getGlobalViewState();

		let body, numRepos = Object.keys(initialState.repos).length, colorVars = '', colorParams = '';
		for (let i = 0; i < initialState.config.graphColours.length; i++) {
			colorVars += '--git-graph-color' + i + ':' + initialState.config.graphColours[i] + '; ';
			colorParams += '[data-color="' + i + '"]{--git-graph-color:var(--git-graph-color' + i + ');} ';
		}

		if (this.dataSource.isGitExecutableUnknown()) {
			body = `<body class="unableToLoad" style="${colorVars}">
			<h2>Unable to load Git Graph</h2>
			<p class="unableToLoadMessage">${UNABLE_TO_FIND_GIT_MSG}</p>
			</body>`;
		} else if (numRepos > 0) {
			body = `<body style="${colorVars}">
			<div id="view">
				<div id="controls">
					<span id="repoControl"><span class="unselectable">Repo: </span><div id="repoDropdown" class="dropdown"></div></span>
					<span id="branchControl"><span class="unselectable">Branches: </span><div id="branchDropdown" class="dropdown"></div></span>
					<label id="showRemoteBranchesControl"><input type="checkbox" id="showRemoteBranchesCheckbox" tabindex="-1"><span class="customCheckbox"></span>Show Remote Branches</label>
					<div id="findBtn" title="Find"></div>
					<div id="settingsBtn" title="Repository Settings"></div>
					<div id="fetchBtn"></div>
					<div id="refreshBtn"></div>
				</div>
				<div id="content">
					<div id="commitGraph"></div>
					<div id="commitTable"></div>
				</div>
				<div id="footer"></div>
			</div>
			<div id="scrollShadow"></div>
			<script nonce="${nonce}">var globalState = ${JSON.stringify(globalState)}, initialState = ${JSON.stringify(initialState)};</script>
			<script src="${this.getMediaUri('out.min.js')}"></script>
			</body>`;
		} else {
			body = `<body class="unableToLoad" style="${colorVars}">
			<h2>Unable to load Git Graph</h2>
			<p class="unableToLoadMessage">No Git repositories were found in the current workspace when it was last scanned by Git Graph.</p>
			<p>If your repositories are in subfolders of the open workspace folder(s), make sure you have set the Git Graph Setting "git-graph.maxDepthOfRepoSearch" appropriately (read the <a href="https://github.com/mhutchie/vscode-git-graph/wiki/Extension-Settings#max-depth-of-repo-search" target="_blank">documentation</a> for more information).</p>
			<p><div id="rescanForReposBtn" class="roundedBtn">Re-scan the current workspace for repositories</div></p>
			<script nonce="${nonce}">(function(){ var api = acquireVsCodeApi(); document.getElementById('rescanForReposBtn').addEventListener('click', function(){ api.postMessage({command: 'rescanForRepos'}); }); })();</script>
			</body>`;
		}
		this.isGraphViewLoaded = numRepos > 0;
		this.loadViewTo = null;

		return `<!DOCTYPE html>
		<html lang="en">
			<head>
				<meta charset="UTF-8">
				<meta http-equiv="Content-Security-Policy" content="default-src 'none'; style-src vscode-resource: 'unsafe-inline'; script-src vscode-resource: 'nonce-${nonce}'; img-src data:;">
				<meta name="viewport" content="width=device-width, initial-scale=1.0">
				<link rel="stylesheet" type="text/css" href="${this.getMediaUri('out.min.css')}">
				<title>Git Graph</title>
				<style>${colorParams}</style>
			</head>
			${body}
		</html>`;
	}

	private getMediaUri(file: string) {
		return this.getUri('media', file).with({ scheme: 'vscode-resource' });
	}

	private getUri(...pathComps: string[]) {
		return vscode.Uri.file(path.join(this.extensionPath, ...pathComps));
	}

	private respondLoadRepos(repos: GitRepoSet, loadViewTo: LoadGitGraphViewTo) {
		this.sendMessage({
			command: 'loadRepos',
			repos: repos,
			lastActiveRepo: this.extensionState.getLastActiveRepo(),
			loadViewTo: loadViewTo
		});
	}
}